# -*- coding: utf-8 -*-
"""
Created on Fri Nov 26 00:06:23 2021

@author: Abhilash
"""

import transformers
from transformers.models.longformer.modeling_tf_longformer import TFLongformerModel,TFLongformerSelfAttention,TFLongformerSelfOutput
import tensorflow as tf
from typing import List, Optional, Tuple, Dict
from transformers import PegasusTokenizer, TFPegasusForConditionalGeneration
from transformers.models.pegasus.configuration_pegasus import PegasusConfig
#from transformers.models.pegasus.modeling_tf_pegasus import shift_tokens_right


class LongformerForPegasus(TFPegasusForConditionalGeneration):
    def __init__(self,config,**kwargs):
        super().__init__(config,**kwargs)
        for j,layer in enumerate(self.model.encoder.layers):
            layer.self_attn=TFLongformerSelfAttention(config=config,layer_id=j)

            
class LongformerPegasusConfig(PegasusConfig):
    def __init__(self, attention_window: List[int] = None, attention_dilation: List[int] = None,initializer_range=None,layer_norm_eps=1e-5,
                 hidden_dropout_prob=1e-5,
                 autoregressive: bool = False, attention_mode: str = 'sliding_chunks',
                 gradient_checkpointing: bool = False, **kwargs):
        super().__init__(**kwargs)
        self.attention_window = attention_window
        self.attention_dilation = attention_dilation
        self.autoregressive = autoregressive
        self.attention_mode = attention_mode
        self.gradient_checkpointing = gradient_checkpointing
        self.initializer_range=initializer_range
        self.layer_norm_eps=layer_norm_eps
        self.hidden_dropout_prob=hidden_dropout_prob
        assert self.attention_mode=='sliding_chunks'
        assert self.attention_mode in ['tvm', 'sliding_chunks', 'n2']            
            
class LongformerSelfAttentionPegasus(tf.keras.layers.Layer):
    def __init__(self,config,layer_id,**kwargs):
        super().__init__(config,layer_id,**kwargs)
        self.self_attention=TFLongformerSelfAttention(config,layer_id,name="self")
        self.outputs=TFLongformerSelfOutput(config,name="outputs")
    def call(self,inputs,training=False):
        (hidden_states,attention_mask,layer_head_mask,is_index_masked,is_index_global_attn,is_global_attn)= inputs
        self_outputs=self.self_attention([hidden_states, attention_mask, layer_head_mask, is_index_masked, is_index_global_attn, is_global_attn],
            training=training,)
        attention_output=self.outputs(self_outputs[0],hidden_states,training=training)
        if len(self_outputs==2):
            ops=(attention_output,)+self_outputs[1:]
        else:
            ops=(attention_output,None)
        return ops
               
               
class LongPegasus():
    def create_long_model(self,save_model,attention_window,max_pos):
        self.base_model='google/pegasus-xsum'
        self.model=TFPegasusForConditionalGeneration.from_pretrained(self.base_model)
        self.tokenizer=PegasusTokenizer.from_pretrained(self.base_model,model_max_length=max_pos)
        self.configuration=LongformerPegasusConfig.from_pretrained(self.base_model)
        self.model.config=self.configuration
        self.configuration.attention_probs_dropout_prob=self.configuration.attention_dropout
        self.configuration.architectures= ['LongformerForPegasus']
        self.tokenizer.model_max_length=max_pos
        self.tokenizer.init_kwargs['model_max_length']=max_pos
        current_max_pos, embed_size = self.model.model.encoder.embed_positions.weight.shape
        assert current_max_pos == self.configuration.max_position_embeddings

        self.configuration.max_encoder_position_embeddings = max_pos 
        self.configuration.max_decoder_position_embeddings = self.configuration.max_position_embeddings #512
        print("max_encoder_position_embeddings: ", self.configuration.max_encoder_position_embeddings)
        print("max_decoder_position_embeddings: ", self.configuration.max_decoder_position_embeddings)
        assert max_pos>current_max_pos
        new_encoder_positional_embedded=self.model.model.encoder.embed_positions.weight
        temp=tf.random.normal(shape=(max_pos,embed_size))
        new_encoder_positional_embed=tf.Variable(temp)
        #new_encoder_positional_embed=tf.tensor(max_pos,embed_size)
        #bart=2
        #pegasus=0
        k=0
        step=current_max_pos-k
        while(k<max_pos-1):
            #step loop over to increase embedding
            new_encoder_positional_embed[k:(k+step)].assign(self.model.model.encoder.embed_positions.weight[:])
            k+=step
        self.model.model.encoder.embed_positions=new_encoder_positional_embed
        self.configuration.attention_window = [attention_window] * self.configuration.num_hidden_layers
        self.configuration.attention_dilation = [1] * self.configuration.num_hidden_layers
        for i, layer in enumerate(self.model.model.encoder.layers):
            longformer_self_attn_for_pegasus = LongformerSelfAttentionPegasus(self.configuration, layer_id=i)
            layer.self_attn = longformer_self_attn_for_pegasus
        print("=======Model Configuration=======")
        print(self.model.config)
        self.model.save_pretrained(save_model)
        self.tokenizer.save_pretrained(save_model)
        print("==============Model & Tokenizer saved===============")
        return self.model, self.tokenizer
        
               
               

 
        



# -*- coding: utf-8 -*-
"""
Created on Fri Nov 26 00:12:08 2021

@author: Abhilash
"""

from LongPegasus import LongPegasus
from transformers import PegasusTokenizer, TFPegasusForConditionalGeneration

if __name__=='__main__':
    l=LongPegasus()             
    model,tokenizer=l.create_long_model(save_model="E:\\Pegasus\\", attention_window=512, max_pos=4096)
    model = TFPegasusForConditionalGeneration.from_pretrained('E:/Pegasus/')
    tokenizer = PegasusTokenizer.from_pretrained('E:/Pegasus/')

    ARTICLE_TO_SUMMARIZE = (
    "PG&E stated it scheduled the blackouts in response to forecasts for high winds "
    "amid dry conditions. The aim is to reduce the risk of wildfires. Nearly 800 thousand customers were "
    "scheduled to be affected by the shutoffs which were expected to last through at least midday tomorrow."
    )
    inputs = tokenizer([ARTICLE_TO_SUMMARIZE], max_length=1024, return_tensors='tf')
    
    # Generate Summary
    summary_ids = model.generate(inputs['input_ids'])
    print([tokenizer.decode(g, skip_special_tokens=True, clean_up_tokenization_spaces=False) for g in summary_ids])


# -*- coding: utf-8 -*-
"""
Created on Tue Mar 16 22:20:30 2021

@author: Abhilash
"""

from distutils.core import setup
setup(
  name = 'LongPegasus',         
  packages = ['LongPegasus'],   
  version = '0.1',       
  license='MIT',        
  description = 'A Longer Version of Pegasus TF Model For Abstractive Summarization',   
  long_description='This package is used for inducing longformer self attention over base pegasus abstractive summarization model to increase the token limit and performance.The Pegasus is a large Transformer-based encoder-decoder model with a new pre-training objective which is adapted to abstractive summarization. More specifically, the pre-training objective, called "Gap Sentence Generation (GSG)", consists of masking important sentences from a document and generating these gap-sentences.On the other hand, the Longformer is a Transformer which replaces the full-attention mechanism (quadratic dependency) with a novel attention mechanism which scale linearly with the input sequence length. Consequently, Longformer can process sequences up to 4,096 tokens long (8 times longer than BERT which is limited to 512 tokens).This package plugs Longformers attention mechanism to Pegasus in order to perform abstractive summarization on long documents. The base modules are built on Tensorflow platform.',
  author = 'ABHILASH MAJUMDER',
  author_email = 'debabhi1396@gmail.com',
  url = 'https://github.com/abhilash1910/LongPegasus',   
  download_url = 'https://github.com/abhilash1910/LongPegasus/archive/v_01.tar.gz',    
  keywords = ['Longformer','Self Attention','Global Attention','Gap sentence generation','Pegasus','Transformer','Encoder Decoder','Tensorflow'],   
  install_requires=[           

          'tensorflow',
          'transformers'
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.7',      
    'Programming Language :: Python :: 3.8',

    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
)
